@@ -8,7 +8,8 @@
 ##' @rdname invest
 ##' @export
 ##' 
-##' @param object An object that inherits from class \code{lm} or \code{nls}.
+##' @param object An object that inherits from class \code{"lm"}, \code{"nls"}, 
+##'               or \code{"lme"}.
 ##' @param y0 The value of the observed response(s) or specified value of the 
 ##'           mean response.
 ##' @param interval The type of interval required.
@@ -20,13 +21,21 @@
 ##'        (\code{TRUE}).
 ##' @param data An optional data frame. This is required if \code{object$data} 
 ##'             is \code{NULL}.
+##' @param boot Logical indicating whether to carry out a bootstrap simulation.
+##' @param nsim Positive integer specifying the number of bootstrap simulations; 
+##'             the bootstrap B (or R).
+##' @param type Character string specifying the type of bootstrap, 
+##'             \code{"parametric"} or \code{"nonparametric"}.
+##' @param seed Optional argument to \code{set.seed}.
+##' @param progress Logical indicating whether to display a text-based progress
+##'                 bar during the bootstrap simulation.
 ##' @param lower The lower endpoint of the interval to be searched.
 ##' @param upper The upper endpoint of the interval to be searched.
 ##' @param q1 Optional lower cutoff to be used in forming confidence intervals. 
-##'           Only used when object inherits from class \code{lme}. Defaults to
+##'           Only used when \code{object} inherits from class \code{"lme"}. Defaults to
 ##'           \code{qnorm((1+level)/2)}.
 ##' @param q2 Optional upper cutoff to be used in forming confidence intervals. 
-##'           Only used when object inherits from class \code{lme}. Defaults to
+##'           Only used when \code{object} inherits from class \code{"lme"}. Defaults to
 ##'           \code{qnorm((1-level)/2)}.
 ##' @param tol The desired accuracy passed on to \code{uniroot}. Recommend a 
 ##'            minimum of 1e-10.
@@ -39,18 +48,30 @@
 ##'          confidence interval. Only needed when \code{adjust = "Bonferroni"}.
 ##' @param ... Additional optional arguments. At present, no optional arguments 
 ##'            are used.
-##' @return An object of class \code{calibrate} containing the following 
-##'         components:
-##' \itemize{
-##'   \item \code{estimate} The estimate of x0.
-##'   \item \code{lwr} The lower confidence limit for x0.
-##'   \item \code{upr} The upper confidence limit for x0.
-##'   \item \code{se} An estimate of the standard error (Wald interval only).
-##'   \item \code{interval} The method used for calculating \code{lower} and 
-##'                   \code{upper} (only used by \code{print} method).
-##' }
+##' @return If \code{boot = FALSE}, then an object of class \code{"calibrate"} 
+##'         containing the following components:
+##'   \itemize{
+##'     \item \code{estimate} The estimate of x0.
+##'     \item \code{lwr} The lower confidence limit for x0.
+##'     \item \code{upr} The upper confidence limit for x0.
+##'     \item \code{se} An estimate of the standard error (Wald interval only).
+##'     \item \code{interval} The method used for calculating \code{lower} and 
+##'           \code{upper} (only used by \code{print} method).
+##'   }
+##'         Otherwise, an object of class \code{"bootCal"} containing the 
+##'         following components:
+##'   \itemize{
+##'     \item \code{original} The estimate of x0.
+##'     \item \code{bootreps} The lower confidence limit for x0.
+##'     \item \code{nsim} The number of bootstrap replicates.
+##'     \item \code{level} The desired confidence level.
+##'   }
 ##' @references
-##' Graybill, F. A., and Iyer, H. K. (1994)
+##' Greenwell, B. M., and Schubert Kabban, C. M. (2014). investr: An R Package 
+##' for Inverse Estimation. \emph{The R Journal}, \bold{6}(1), 90--100. 
+##' URL http://journal.r-project.org/archive/2014-1/greenwell-kabban.pdf.
+##'
+##' Graybill, F. A., and Iyer, H. K. (1994).
 ##' \emph{Regression analysis: Concepts and Applications}. Duxbury Press. 
 ##'
 ##' Huet, S., Bouvier, A., Poursat, M-A., and Jolivet, E.  (2004)
@@ -64,13 +85,24 @@
 ##' \emph{Biometrics} \bold{85}(2): 439--449.
 ##' doi:10.1093/biomet/85.2.439.
 ##' @examples
-##' data(Puromycin, package = "datasets")
-##' Puromycin2 <- Puromycin[Puromycin$state == "treated", ]
-##' Puro2.nls <- nls(rate ~ (theta1 * conc) / (theta2 + conc), 
-##'                  data = Puromycin2, start = list(theta1 = 200, theta2 = 1))
-##' plotFit(Puro2.nls, interval = "both")
-##' invest(Puro2.nls, y0 = 100, interval = "inversion")
-##' invest(Puro2.nls, y0 = 100, interval = "inversion", mean.response = TRUE)
+##' ##
+##' ## Nasturtium example (nonlinear regression with replication)
+##' ##
+##' 
+##' ## Log-logistic model
+##' mod <- nls(weight ~ theta1/(1 + exp(theta2 + theta3 * log(conc))),
+##'            start = list(theta1 = 1000, theta2 = -1, theta3 = 1),
+##'            data = nasturtium)
+##'            
+##' ## Compute approximate 95% calibration intervals
+##' invest(mod, y0 = c(309, 296, 419), interval = "inversion")
+##' invest(mod, y0 = c(309, 296, 419), interval = "Wald")  
+##' 
+##' ## Bootstrap calibration intervals
+##' #boo <- invest(mod, y0 = c(309, 296, 419), boot = TRUE, nsim = 9999, 
+##' #              progress = TRUE, seed = 101)
+##' #boo  # print bootstrap summary
+##' #plot(boo)  # plot results
 invest <- function(object, ...) {
   UseMethod("invest")
 } 
@@ -79,8 +111,10 @@ invest <- function(object, ...) {
 ##' @export
 ##' @method invest lm
 invest.lm <- function(object, y0, interval = c("inversion", "Wald", "none"), 
-                      level = 0.95, mean.response = FALSE, data, lower, upper, 
-                      tol = .Machine$double.eps^0.25, maxiter = 1000,  
+                      level = 0.95, mean.response = FALSE, data, boot = FALSE,
+                      type = c("parametric", "nonparametric"), nsim = 1, 
+                      seed = NULL, progress = FALSE, lower, upper, 
+                      tol = .Machine$double.eps^0.25, maxiter = 1000, 
                       adjust = c("none", "Bonferroni"), k,  ...) {
   
   ## Extract data, variable names, etc.
@@ -120,6 +154,107 @@ invest.lm <- function(object, y0, interval = c("inversion", "Wald", "none"),
          call. = FALSE)
   }
   
+  ## Parametric bootstrap ------------------------------------------------------
+  if (boot) {
+    
+    ## Sanity check
+    stopifnot((nsim <- as.integer(nsim[1])) > 0)
+    
+    ## Set up progress bar (if requested)
+    if (progress) { 
+      pb <- txtProgressBar(min = 0, max = nsim, style = 3)
+    }
+    
+    ## Initialize random number generator
+    if (!exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) 
+      runif(1)
+    if (is.null(seed)) 
+      RNGstate <- get(".Random.seed", envir = .GlobalEnv)
+    else {
+      R.seed <- get(".Random.seed", envir = .GlobalEnv)
+      set.seed(seed)
+      RNGstate <- structure(seed, kind = as.list(RNGkind()))
+      on.exit(assign(".Random.seed", R.seed, envir = .GlobalEnv))
+    }
+    
+    ## Simulate new response vectors
+    ftd <- fitted(object)  # fitted values
+    res <- residuals(object) # redisuals
+    type <- match.arg(type)
+    if (type == "parametric") {  
+      ss <- simulate(object, nsim = nsim)
+    } else {
+      ss <- replicate(nsim, ftd + sample(res, replace = TRUE), simplify = FALSE)
+    }
+    
+    ## Function to calculate inverse estimate
+    x0Fun <- function(i) {
+      
+      ## Update model using simulated response data
+      boot.data <- eval(object$call$data)  # copy data
+      boot.data[, yname] <- ss[[i]]  # simulated response vector
+      boot.object <- tryCatch(update(object, data = boot.data),
+                              error = function(e) NULL)
+      
+      ## If updating the model fails, then return value is NA
+      if (is.null(boot.object)) {
+        ret <- NA
+      } else {
+        
+        ## Simulate new response (different from simulated response vector)
+        if (mean.response) {  # regulation
+          y0.star <- y0  # hold constant in bootstrap replications
+        } else {  # calibration
+          if (type == "parametric") {
+            y0.star <- y0 + rnorm(length(y0), sd = Sigma(object))
+          } else {
+            y0.star <- y0 + sample(res, size = length(y0), replace = TRUE)
+          }
+        }
+        
+        ## Calculate point estimate
+        ret <- tryCatch(uniroot(function(x) {
+          predict(boot.object, newdata = makeData(x, xname)) - mean(y0.star)
+        }, interval = c(lower, upper), tol = tol, maxiter = maxiter)$root, 
+        error = function(e) NA)
+      }
+      
+      ## Update progress bar
+      if (progress) { 
+        setTxtProgressBar(pb, i) 
+      }
+      
+      ## Return estimate
+      ret
+      
+    }
+    
+    ## Calculate bootstrap replicates
+    x0.star <- sapply(seq_len(nsim), x0Fun)
+    
+    ## Check for errors and return the runs that did not fail
+    if (anyNA(x0.star)) {
+      num_fail <- sum(is.na(x0.star))
+      warning("some bootstrap runs failed (", num_fail, "/", nsim, 
+              ")")
+      x0.star <- na.omit(x0.star)  # remove runs that failed
+      attributes(x0.star) <- NULL  # remove attributes
+    } else {
+      num_fail <- NULL
+    }
+    
+    ## Create and return a bootCal object (essentially a list that can later be 
+    ## converted to an object of class boot)
+    res <- list(original  = x0.est,  # original estimate
+                bootreps  = x0.star,  # bootstrap replicates
+                nsim      = nsim,  # number of simulations
+                level     = level)  # desired confidence level
+    class(res) = "bootCal"
+    attr(res, "bootFail") <- num_fail
+    return(res)
+     
+  }
+  
   ## Return point estimate only
   interval <- match.arg(interval)
   if (interval == "none") return(x0.est)
@@ -225,8 +360,10 @@ invest.lm <- function(object, y0, interval = c("inversion", "Wald", "none"),
 ##' @export
 ##' @method invest nls
 invest.nls <- function(object, y0, interval = c("inversion", "Wald", "none"),  
-                       level = 0.95, mean.response = FALSE, data, lower, upper, 
-                       tol = .Machine$double.eps^0.25, maxiter = 1000,
+                       level = 0.95, mean.response = FALSE, data, boot = FALSE,
+                       type = c("parametric", "nonparametric"), nsim = 1, 
+                       seed = NULL, progress = FALSE, lower, upper, 
+                       tol = .Machine$double.eps^0.25, maxiter = 1000, 
                        adjust = c("none", "Bonferroni"), k, ...) {
   
   ## TODO:
@@ -279,6 +416,107 @@ invest.nls <- function(object, y0, interval = c("inversion", "Wald", "none"),
          call. = FALSE)
   }
   
+  ## Parametric bootstrap ------------------------------------------------------
+  if (boot) {
+    
+    ## Sanity check
+    stopifnot((nsim <- as.integer(nsim[1])) > 0)
+    
+    ## Set up progress bar (if requested)
+    if (progress) { 
+      pb <- txtProgressBar(min = 0, max = nsim, style = 3)
+    }
+    
+    ## Initialize random number generator
+    if (!exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) 
+      runif(1)
+    if (is.null(seed)) 
+      RNGstate <- get(".Random.seed", envir = .GlobalEnv)
+    else {
+      R.seed <- get(".Random.seed", envir = .GlobalEnv)
+      set.seed(seed)
+      RNGstate <- structure(seed, kind = as.list(RNGkind()))
+      on.exit(assign(".Random.seed", R.seed, envir = .GlobalEnv))
+    }
+    
+    ## Simulate new response vectors
+    ftd <- fitted(object)  # fitted values
+    res <- residuals(object) # redisuals
+    type <- match.arg(type)
+    if (type == "parametric") {  
+      ss <- simulate(object, nsim = nsim)
+    } else {
+      ss <- replicate(nsim, ftd + sample(res, replace = TRUE), simplify = FALSE)
+    }
+    
+    ## Function to calculate inverse estimate
+    x0Fun <- function(i) {
+      
+      ## Update model using simulated response data
+      boot.data <- eval(object$call$data)  # copy data
+      boot.data[, yname] <- ss[[i]]  # simulated response vector
+      boot.object <- tryCatch(update(object, data = boot.data),
+                              error = function(e) NULL)
+      
+      ## If updating the model fails, then return value is NA
+      if (is.null(boot.object)) {
+        ret <- NA
+      } else {
+        
+        ## Simulate new response (different from simulated response vector)
+        if (mean.response) {  # regulation
+          y0.star <- y0  # hold constant in bootstrap replications
+        } else {  # calibration
+          if (type == "parametric") {
+            y0.star <- y0 + rnorm(length(y0), sd = Sigma(object))
+          } else {
+            y0.star <- y0 + sample(res, size = length(y0), replace = TRUE)
+          }
+        }
+        
+        ## Calculate point estimate
+        ret <- tryCatch(uniroot(function(x) {
+            predict(boot.object, newdata = makeData(x, xname)) - mean(y0.star)
+          }, interval = c(lower, upper), tol = tol, maxiter = maxiter)$root, 
+          error = function(e) NA)
+      }
+      
+      ## Update progress bar
+      if (progress) { 
+        setTxtProgressBar(pb, i) 
+      }
+      
+      ## Return estimate
+      ret
+      
+    }
+    
+    ## Calculate bootstrap replicates
+    x0.star <- sapply(seq_len(nsim), x0Fun)
+    
+    ## Check for errors and return the runs that did not fail
+    if (anyNA(x0.star)) {
+      num_fail <- sum(is.na(x0.star))
+      warning("some bootstrap runs failed (", num_fail, "/", nsim, 
+              ")")
+      x0.star <- na.omit(x0.star)  # remove runs that failed
+      attributes(x0.star) <- NULL  # remove attributes
+    } else {
+      num_fail <- NULL
+    }
+
+    ## Create and return a bootCal object (essentially a list that can later be 
+    ## converted to an object of class boot)
+    res <- list(original  = x0.est,  # original estimate
+                bootreps  = x0.star,  # bootstrap replicates
+                nsim      = nsim,  # number of simulations
+                level     = level)  # desired confidence level
+    class(res) = "bootCal"
+    attr(res, "bootFail") <- num_fail
+    return(res)
+
+  }
+  
   ## Return point estimate only
   interval <- match.arg(interval)
   if (interval == "none") return(x0.est)
@@ -518,3 +756,60 @@ invest.lme <- function(object, y0, interval = c("inversion", "Wald", "none"),
   res
   
 }
+
+##' @keywords internal
+print.bootCal <- function(x, digits = 4, ...) { 
+  ci <- round(quantile(x$bootreps, probs = c((1-x$level)/2, (1+x$level)/2)),
+              digits = digits)
+  names(ci) <- c("lower", "upper")
+  op <- c("estimate" = x$original, 
+          "se"       = sd(x$bootreps), 
+          "bias"     = mean(x$bootreps) - x$original)
+  print(round(op, digits = digits))
+  cat("\n", "Percentile bootstrap interval: (", ci[1], ",", ci[2], ")", "\n")
+  invisible(x)
+} 
+
+##' Plots of the Output of a Bootstrap Calibration Simulation
+##' 
+##' This takes a bootstrap calibration object and produces plots for the 
+##' bootstrap replicates of the inverse estimate.
+##' 
+##' @rdname plot.bootCal
+##' @export
+##' @method plot bootCal
+##' 
+##' @param x An object that inherits from class \code{"bootCal"}.
+##' @param ... Additional optional arguments. At present, no optional arguments 
+##'            are used.
+plot.bootCal <- function(x, ...) {
+  
+  t <- x$bootreps  # bootstrap replicates
+  t0 <- x$original  # original estimate
+  
+  ## Calculate number of histogram breaks (better than default)
+  if (!is.null(t0)) {
+    nclass <- min(max(ceiling(length(t)/25), 10), 100)
+    rg <- range(t)
+    if (t0 < rg[1L]) 
+      rg[1L] <- t0
+    else if (t0 > rg[2L]) 
+      rg[2L] <- t0
+    rg <- rg + 0.05 * c(-1, 1) * diff(rg)
+    lc <- diff(rg)/(nclass - 2)
+    n1 <- ceiling((t0 - rg[1L])/lc)
+    n2 <- ceiling((rg[2L] - t0)/lc)
+    bks <- t0 + (-n1:n2) * lc
+  }
+  
+  ## Plots
+  par(mfrow = c(1, 2))
+  hist(t, breaks = bks, probability = TRUE, main = "",
+       xlab = "Bootstrap replicate")
+  qqnorm(t, xlab = "Standard normal quantile", ylab = "Bootstrap quantile",
+         main = "")
+  qqline(t)
+  par(mfrow = c(1, 1))
+  invisible(x)
+  
+}
\ No newline at end of file
